
//---------------------------------------------------------------------
//---------------------------------------------------------------------
//
//  header.h
//
//---------------------------------------------------------------------
//---------------------------------------------------------------------
#ifndef __HEADER_H
#define __HEADER_H

//---------------------------------------------------------------------
// The following include file is generated automatically by the
// "setparams" utility. It defines 
//      maxcells:      the square root of the maximum number of processors
//      problem_size:  12, 64, 102, 162 (for class T, A, B, C)
//      dt_default:    default time step for this problem size if no
//                     config file
//      niter_default: default number of iterations for this problem size
//---------------------------------------------------------------------

#include <upc_relaxed.h>
#include "npbparams.h"
#include "../common/npb-C.h"

#define EAST   2000
#define WEST   3000
#define NORTH  4000
#define SOUTH  5000
#define BOTTOM 6000
#define TOP    7000

#define MAX_CELL_DIM ((PROBLEM_SIZE/MAXCELLS)+1)
#define IMAX MAX_CELL_DIM
#define JMAX MAX_CELL_DIM
#define KMAX MAX_CELL_DIM

#define IMAXP (IMAX/2*2+1)
#define JMAXP (JMAX/2*2+1)

//---------------------------------------------------------------------
// +1 at end to avoid zero length arrays for 1 node
//---------------------------------------------------------------------
#define BUF_SIZE (MAX_CELL_DIM*MAX_CELL_DIM*(MAXCELLS-1)*60*2+1)
#define BUF_SIZZ ((2*MAXCELLS-2)/MAXCELLS*MAX_CELL_DIM*MAX_CELL_DIM*22+1)

#define T_TOTAL  1
#define T_RHS    2
#define T_XSOLVE 3
#define T_YSOLVE 4
#define T_ZSOLVE 5
#define T_BPACK  6
#define T_EXCH   7
#define T_XCOMM  8
#define T_YCOMM  9
#define T_ZCOMM 10
#define T_LAST  10

#define USE_MTIMERS
#ifdef NO_MTIMERS
#undef USE_MTIMERS
#define TIMER_START(x)
#define TIMER_STOP(x)
#else
#define TIMER_START(x) if (timeron) timer_start(x)
#define TIMER_STOP(x) if (timeron) timer_stop(x)
#endif

#define SQR(x) (x)*(x)
#define DMAX1(x,y) ((x)>(y)? (x):(y))
#define DMAX4(x,y,z,p) DMAX1(DMAX1(DMAX1(x,y),z),p)

#define grid_points(i) grid_points[i-1]
#define ce(i,j) ce[(i-1)+(5)*(j-1)]
#define cell_coord(i,j) cell_coord[(i-1)+(3)*(j-1)]
#define cell_low(i,j) cell_low[(i-1)+(3)*(j-1)]
#define cell_high(i,j) cell_high[(i-1)+(3)*(j-1)]
#define cell_size(i,j) cell_size[(i-1)+(3)*(j-1)]
#define predecessor(i) predecessor[i-1]
#define slice(i,j) slice[(i-1)+(3)*(j-1)]
#define grid_size(i) grid_size[i-1]
#define successor(i) successor[i-1]
#define start(i,j) start[(i-1)+(3)*(j-1)]
#define end(i,j) end[(i-1)+(3)*(j-1)]
#define u(i,j,k,m,n) u[(i+2)+(IMAXP+4)*((j+2)+(JMAXP+4)*((k+2)+(KMAX+4)*((m-1)+(5)*(n-1))))]
#define us(i,j,k,m) us[(i+1)+(IMAX+2)*((j+1)+(JMAX+2)*((k+1)+(KMAX+2)*(m-1)))]
#define vs(i,j,k,m) vs[(i+1)+(IMAX+2)*((j+1)+(JMAX+2)*((k+1)+(KMAX+2)*(m-1)))]
#define ws(i,j,k,m) ws[(i+1)+(IMAX+2)*((j+1)+(JMAX+2)*((k+1)+(KMAX+2)*(m-1)))]
#define qs(i,j,k,m) qs[(i+1)+(IMAX+2)*((j+1)+(JMAX+2)*((k+1)+(KMAX+2)*(m-1)))]
#define ainv(i,j,k,m) ainv[(i+1)+(IMAX+2)*((j+1)+(JMAX+2)*((k+1)+(KMAX+2)*(m-1)))]
#define rho_i(i,j,k,m) rho_i[(i+1)+(IMAX+2)*((j+1)+(JMAX+2)*((k+1)+(KMAX+2)*(m-1)))]
#define speed(i,j,k,m) speed[(i+1)+(IMAX+2)*((j+1)+(JMAX+2)*((k+1)+(KMAX+2)*(m-1)))]
#define square(i,j,k,m) square[(i+1)+(IMAX+2)*((j+1)+(JMAX+2)*((k+1)+(KMAX+2)*(m-1)))]
#define rhs(i,j,k,m,n) rhs[(i)+(IMAXP)*((j)+(JMAXP)*((k)+(KMAX)*((m-1)+(5)*(n-1))))]
#define forcing(i,j,k,m,n) forcing[(i)+(IMAXP)*((j)+(JMAXP)*((k)+(KMAX)*((m-1)+(5)*(n-1))))]
#define lhs(i,j,k,m,n) lhs[(i)+(IMAXP)*((j)+(JMAXP)*((k)+(KMAX)*((m-1)+(15)*(n-1))))]
#define in_buffer(i) in_buffer[i-1]
#define out_buffer(i) out_buffer[i-1]
#define cv(i) cv[i+2]
#define rhon(i) rhon[i+2]
#define rhos(i) rhos[i+2]
#define rhoq(i) rhoq[i+2]
#define cuf(i) cuf[i+2]
#define q(i) q[i+2]
#define ue(i,j) ue[(i+2)+(MAX_CELL_DIM+4)*(j-1)]
#define buf(i,j) buf[(i+2)+(MAX_CELL_DIM+4)*(j-1)]

#ifdef G_MAIN
      int           ncells, grid_points[3];
//      common /global/   ncells, grid_points;

      double tx1, tx2, tx3, ty1, ty2, ty3, tz1, tz2, tz3, 
                        dx1, dx2, dx3, dx4, dx5, dy1, dy2, dy3, dy4, 
                        dy5, dz1, dz2, dz3, dz4, dz5, dssp, dt, 
                        ce[5*13], dxmax, dymax, dzmax, xxcon1, xxcon2, 
                        xxcon3, xxcon4, xxcon5, dx1tx1, dx2tx1, dx3tx1,
                        dx4tx1, dx5tx1, yycon1, yycon2, yycon3, yycon4,
                        yycon5, dy1ty1, dy2ty1, dy3ty1, dy4ty1, dy5ty1,
                        zzcon1, zzcon2, zzcon3, zzcon4, zzcon5, dz1tz1, 
                        dz2tz1, dz3tz1, dz4tz1, dz5tz1, dnxm1, dnym1, 
                        dnzm1, c1c2, c1c5, c3c4, c1345, conz1, c1, c2, 
                        c3, c4, c5, c4dssp, c5dssp, dtdssp, dttx1, bt,
                        dttx2, dtty1, dtty2, dttz1, dttz2, c2dttx1, 
                        c2dtty1, c2dttz1, comz1, comz4, comz5, comz6, 
                        c3c4tx3, c3c4ty3, c3c4tz3, c2iv, con43, con16;

//      common /constants/ tx1, tx2, tx3, ty1, ty2, ty3, tz1, tz2, tz3,
//                        dx1, dx2, dx3, dx4, dx5, dy1, dy2, dy3, dy4, 
//                        dy5, dz1, dz2, dz3, dz4, dz5, dssp, dt, 
//                        ce, dxmax, dymax, dzmax, xxcon1, xxcon2, 
//                        xxcon3, xxcon4, xxcon5, dx1tx1, dx2tx1, dx3tx1,
//                        dx4tx1, dx5tx1, yycon1, yycon2, yycon3, yycon4,
//                        yycon5, dy1ty1, dy2ty1, dy3ty1, dy4ty1, dy5ty1,
//                        zzcon1, zzcon2, zzcon3, zzcon4, zzcon5, dz1tz1, 
//                        dz2tz1, dz3tz1, dz4tz1, dz5tz1, dnxm1, dnym1, 
//                        dnzm1, c1c2, c1c5, c3c4, c1345, conz1, c1, c2, 
//                        c3, c4, c5, c4dssp, c5dssp, dtdssp, dttx1, bt,
//                        dttx2, dtty1, dtty2, dttz1, dttz2, c2dttx1, 
//                        c2dtty1, c2dttz1, comz1, comz4, comz5, comz6, 
//                        c3c4tx3, c3c4ty3, c3c4tz3, c2iv, con43, con16;

      int     cell_coord[3*MAXCELLS], cell_low[3*MAXCELLS], 
              cell_high[3*MAXCELLS], cell_size[3*MAXCELLS],
              predecessor[3],        slice[3*MAXCELLS],
              grid_size[3],          successor[3],
              start[3*MAXCELLS],     end[3*MAXCELLS];
//      common /partition/ cell_coord, cell_low, cell_high, cell_size,
//                         grid_size, successor, predecessor, slice,
//                         start, end;


      double 
         u[(IMAXP+4)*(JMAXP+4)*(KMAX+4)*5*MAXCELLS],
         us[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         vs[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         ws[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         qs[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         ainv[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         rho_i[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         speed[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         square[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         rhs[(IMAXP)*(JMAXP)*(KMAX)*5*MAXCELLS],
         forcing[(IMAXP)*(JMAXP)*(KMAX)*5*MAXCELLS],
         lhs[(IMAXP)*(JMAXP)*(KMAX)*15*MAXCELLS],
         in_buffer[BUF_SIZE], out_buffer[BUF_SIZE];
//      common /fields/  u, us, vs, ws, qs, ainv, rho_i, speed, square, 
//                       rhs, forcing, lhs, in_buffer, out_buffer;

      double cv[MAX_CELL_DIM+4],   rhon[MAX_CELL_DIM+4],
                       rhos[MAX_CELL_DIM+4], rhoq[MAX_CELL_DIM+4],
                       cuf[MAX_CELL_DIM+4],  q[MAX_CELL_DIM+4],
                       ue[(MAX_CELL_DIM+4)*5], buf[(MAX_CELL_DIM+4)*5];
//      common /work_1d/ cv, rhon, rhos, rhoq, cuf, q, ue, buf;

      int  west_size, east_size, bottom_size, top_size,
               north_size, south_size, start_send_west, 
               start_send_east, start_send_south, start_send_north,
               start_send_bottom, start_send_top, start_recv_west,
               start_recv_east, start_recv_south, start_recv_north,
               start_recv_bottom, start_recv_top;
//      common /box/ west_size, east_size, bottom_size,
//                   top_size, north_size, south_size, 
//                   start_send_west, start_send_east, start_send_south,
//                   start_send_north, start_send_bottom, start_send_top,
//                   start_recv_west, start_recv_east, start_recv_south,
//                   start_recv_north, start_recv_bottom, start_recv_top;

      int timeron;
//      common /tflags/ timeron

#else
extern int           ncells, grid_points[3];

extern double tx1, tx2, tx3, ty1, ty2, ty3, tz1, tz2, tz3, 
                        dx1, dx2, dx3, dx4, dx5, dy1, dy2, dy3, dy4, 
                        dy5, dz1, dz2, dz3, dz4, dz5, dssp, dt, 
                        ce[5*13], dxmax, dymax, dzmax, xxcon1, xxcon2, 
                        xxcon3, xxcon4, xxcon5, dx1tx1, dx2tx1, dx3tx1,
                        dx4tx1, dx5tx1, yycon1, yycon2, yycon3, yycon4,
                        yycon5, dy1ty1, dy2ty1, dy3ty1, dy4ty1, dy5ty1,
                        zzcon1, zzcon2, zzcon3, zzcon4, zzcon5, dz1tz1, 
                        dz2tz1, dz3tz1, dz4tz1, dz5tz1, dnxm1, dnym1, 
                        dnzm1, c1c2, c1c5, c3c4, c1345, conz1, c1, c2, 
                        c3, c4, c5, c4dssp, c5dssp, dtdssp, dttx1, bt,
                        dttx2, dtty1, dtty2, dttz1, dttz2, c2dttx1, 
                        c2dtty1, c2dttz1, comz1, comz4, comz5, comz6, 
                        c3c4tx3, c3c4ty3, c3c4tz3, c2iv, con43, con16;

extern int     cell_coord[3*MAXCELLS], cell_low[3*MAXCELLS], 
              cell_high[3*MAXCELLS], cell_size[3*MAXCELLS],
              predecessor[3],        slice[3*MAXCELLS],
              grid_size[3],          successor[3],
              start[3*MAXCELLS],     end[3*MAXCELLS];

extern double 
         u[(IMAXP+4)*(JMAXP+4)*(KMAX+4)*5*MAXCELLS],
         us[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         vs[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         ws[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         qs[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         ainv[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         rho_i[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         speed[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         square[(IMAX+2)*(JMAX+2)*(KMAX+2)*MAXCELLS],
         rhs[(IMAXP)*(JMAXP)*(KMAX)*5*MAXCELLS],
         forcing[(IMAXP)*(JMAXP)*(KMAX)*5*MAXCELLS],
         lhs[(IMAXP)*(JMAXP)*(KMAX)*15*MAXCELLS],
         in_buffer[BUF_SIZE], out_buffer[BUF_SIZE];

extern double cv[MAX_CELL_DIM+4],   rhon[MAX_CELL_DIM+4],
                       rhos[MAX_CELL_DIM+4], rhoq[MAX_CELL_DIM+4],
                       cuf[MAX_CELL_DIM+4],  q[MAX_CELL_DIM+4],
                       ue[(MAX_CELL_DIM+4)*5], buf[(MAX_CELL_DIM+4)*5];

extern int  west_size, east_size, bottom_size, top_size,
               north_size, south_size, start_send_west, 
               start_send_east, start_send_south, start_send_north,
               start_send_bottom, start_send_top, start_recv_west,
               start_recv_east, start_recv_south, start_recv_north,
               start_recv_bottom, start_recv_top;

extern int timeron;

#endif /*G_MAIN*/

extern void exact_solution(double xi,double eta,double zeta,double dtemp[]);

extern void setup_upc(int *argc, char ***argv);
extern void set_constants(void);
extern void compute_buffer_size(int dim);
extern void make_set(void);
extern void initialize(void);
extern void lhsinit(void);
extern void exact_rhs(void);
extern void adi(void);
extern void compute_rhs(void);
extern void copy_faces(void);
extern void x_solve(void);
extern void y_solve(void);
extern void z_solve(void);
extern void add(void);
extern void verify(int niter, char *class, int *verified);
extern void error_norm(double rms[]);
extern void rhs_norm(double rms[]);
extern void ninvr(int c);
extern void pinvr(int c);
extern void txinvr(void);
extern void tzetar(int c);
extern void lhsx(int c);
extern void lhsy(int c);
extern void lhsz(int c);

#endif
