//-------------------------------------------------------------------------!
//                                                                         !
//        N  A  S     P A R A L L E L     B E N C H M A R K S  3.3         !
//                                                                         !
//                                   B T                                   !
//                                                                         !
//-------------------------------------------------------------------------!
//                                                                         !
//    This benchmark is part of the NAS Parallel Benchmark 3.3 suite.      !
//    It is described in NAS Technical Reports 95-020 and 02-007.          !
//                                                                         !
//    Permission to use, copy, distribute and modify this software         !
//    for any purpose with or without fee is hereby granted.  We           !
//    request, however, that all derived work reference the NAS            !
//    Parallel Benchmarks 3.3. This software is provided "as is"           !
//    without express or implied warranty.                                 !
//                                                                         !
//    Information on NPB 3.3, including the technical report, the          !
//    original specifications, source code, results and information        !
//    on how to submit new results, is available at:                       !
//                                                                         !
//           http://www.nas.nasa.gov/Software/NPB/                         !
//                                                                         !
//    Send comments or suggestions to  npb@nas.nasa.gov                    !
//                                                                         !
//          NAS Parallel Benchmarks Group                                  !
//          NASA Ames Research Center                                      !
//          Mail Stop: T27A-1                                              !
//          Moffett Field, CA   94035-1000                                 !
//                                                                         !
//          E-mail:  npb@nas.nasa.gov                                      !
//          Fax:     (650) 604-3957                                        !
//                                                                         !
//-------------------------------------------------------------------------!

//---------------------------------------------------------------------
//
// Authors: R. F. Van der Wijngaart
//          T. Harris
//          M. Yarrow
//          H. Jin  (UPC & UPC-IO version)
//
//---------------------------------------------------------------------
#include <stdio.h>
#define G_MAIN
#include "header.h"
#include "upcnpb.h"
#include <unistd.h>

#define BSIZE 132

shared [] double sh_rbuf[2];
shared [] int sh_ibuf[8];

//---------------------------------------------------------------------
//      program MPBT;
//---------------------------------------------------------------------
int main(int argc, char *argv[]) {
      
       int niter, step, c;
       double navg, mflops, mbytes, n3;

       double t, tmax, tiominv, tpc;
       int verified;
       char class;

       int wr_interval;

       FILE *fp;
       char cbuf[BSIZE];

  unsigned int mymask = 1 << (MYTHREAD%16);
  //sched_setaffinity(0, sizeof(unsigned int), &mymask);
       setup_upc(&argc, &argv);
       if (!active) goto done;

//---------------------------------------------------------------------
//      Root node reads input file (if it exists) else takes
//      defaults from parameters
//---------------------------------------------------------------------
       if (node == root) {
          
          printf("\n\n NAS Parallel Benchmarks 3.3 -- BT Benchmark\n");
          fp = fopen("inputbt.data", "r");

          rd_interval = 0;
          if (fp != NULL) {
            printf(" Reading from input file inputbt.data\n");
            fgets(cbuf, BSIZE, fp);
            sscanf(cbuf, "%d", &niter);
            fgets(cbuf, BSIZE, fp);
            sscanf(cbuf, "%lf", &dt);
            fgets(cbuf, BSIZE, fp);
            sscanf(cbuf, "%d %d %d", 
                   &grid_points(1), &grid_points(2), &grid_points(3));
            if (iotype != 0) {
                fgets(cbuf, BSIZE, fp);
                if (sscanf(cbuf, "%d %d", &wr_interval, &rd_interval) < 2)
                    rd_interval = 0;
                if (wr_interval <= 0) wr_interval = wr_default;
            }
            if (iotype == 1) {
                fgets(cbuf, BSIZE, fp);
                sscanf(cbuf, "%d %d", &collbuf_nodes, &collbuf_size);
                printf(" collbuf_nodes %d\n", collbuf_nodes);
                printf(" collbuf_size %d\n", collbuf_size);
            }
            fclose(fp);
          } else {
            printf(" No input file inputbt.data. Using compiled defaults\n");
            niter = NITER_DEFAULT;
            dt    = dt_default;
            grid_points(1) = PROBLEM_SIZE;
            grid_points(2) = PROBLEM_SIZE;
            grid_points(3) = PROBLEM_SIZE;
            wr_interval = wr_default;
            if (iotype == 1) {
//             set number of nodes involved in collective buffering to 4,
//             unless total number of nodes is smaller than that.
//             set buffer size for collective buffering to 1MB per node
//             collbuf_nodes = min(4,no_nodes)
//             set default to No-File-Hints with a value of 0
              collbuf_nodes = 0;
              collbuf_size = 1000000;
            }

          }

          printf(" Size: %4dx%4dx%4d\n", 
                 grid_points(1), grid_points(2), grid_points(3));
          printf(" Iterations: %4d    dt: %11.7f\n", niter, dt);
          if (no_nodes != total_nodes)
              printf(" Total number of processes: %5d\n", total_nodes);
          if (no_nodes != MAXCELLS*MAXCELLS) 
              printf(" WARNING: compiled for %5d processes\n",
                     MAXCELLS*MAXCELLS);
          printf(" Number of active processes: %5d\n\n", no_nodes);

          if (iotype != 0) {
              printf(" BTIO -- %s write interval: %3d\n\n",
                     (iotype==1)? "FULL UPC-IO" :
                     (iotype==2)? "SIMPLE UPC-IO" :
                     (iotype==3)? "EPIO" : "POSIX IO", wr_interval);
          }

       }


//  ... broadcast parameters

       if (node == root && no_nodes > 1) {
          sh_ibuf[0] = niter;
          sh_ibuf[1] = grid_points(1);
          sh_ibuf[2] = grid_points(2);
          sh_ibuf[3] = grid_points(3);
          sh_ibuf[4] = rd_interval;
          sh_ibuf[5] = wr_interval;
          sh_ibuf[6] = collbuf_nodes;
          sh_ibuf[7] = collbuf_size;
          sh_rbuf[0] = dt;
       }
       upc_barrier;
       if (node != root) {
          niter          = sh_ibuf[0];
          grid_points(1) = sh_ibuf[1];
          grid_points(2) = sh_ibuf[2];
          grid_points(3) = sh_ibuf[3];
          rd_interval    = sh_ibuf[4];
          wr_interval    = sh_ibuf[5];
          collbuf_nodes  = sh_ibuf[6];
          collbuf_size   = sh_ibuf[7];
          dt             = sh_rbuf[0];
       }


       make_set();

       for (c = 1; c <= MAXCELLS; c++) {
          if ( (cell_size(1,c) > IMAX) ||
               (cell_size(2,c) > JMAX) ||
               (cell_size(3,c) > KMAX) ) {
             printf(" %d %d %d %d %d\n", node, c, cell_size(1,c),
                     cell_size(2,c), cell_size(3,c));
             printf(" Problem size too big for compiled array sizes\n");
             goto done;
          }
       }

       set_constants();

       initialize();

       setup_btio();
       idump = 0;

       lhsinit();

       exact_rhs();

       compute_buffer_size(5);

//---------------------------------------------------------------------
//      do one time step to touch all code, and reinitialize
//---------------------------------------------------------------------
       adi();
       initialize();

       timer_clear(2);
       timer_clear(1);

//---------------------------------------------------------------------
//      Synchronize before placing time stamp
//---------------------------------------------------------------------
       upc_barrier;

       timer_start(1);

       for (step = 1; step <= niter; step++) {

          if (node == root) {
             if ((step%20) == 0 || step == niter ||
                 step == 1) {
                printf(" Time step %4d\n", step);
             }
          }

          adi();

          if (iotype != 0) {
              if ((step%wr_interval)==0 || step == niter) {
                  if (node == root) {
                      printf(" Writing data set, time step %4d\n", step);
                  }
                  if (step == niter && rd_interval > 1) {
                      rd_interval = 1;
                  }
                  timer_start(2);
                  output_timestep();
                  timer_stop(2);
                  idump = idump + 1;
              }
          }
       }

       timer_start(2);
       btio_cleanup();
       timer_stop(2);
       timer_stop(1);
       t = timer_read(1);
       
       verify(niter, &class, &verified);

       mupc_reduce_r8(&t, &tmax, 1, SMP_OPR_MAX, root);

       if (iotype != 0) {
          t = timer_read(2);
          if (t != 0.e0) t = 1.0e0 / t;
          mupc_reduce_r8(&t, &tiominv, 1, SMP_OPR_SUM, root);
       }
      double wtime;

       if( node == root ) {
          n3 = 1.0e0*grid_points(1)*grid_points(2)*grid_points(3);
          navg = (grid_points(1)+grid_points(2)+grid_points(3))/3.0;
          if( tmax != 0. ) {
             mflops = 1.0e-6*(double)(niter)*
               (3478.8*(double)n3-17655.7*navg*navg+28023.7*navg)
               / tmax;
          } else {
             mflops = 0.0;
          }

          if (iotype != 0) {
             mbytes = n3 * 40.0 * idump * 1.0e-6;
             tiominv = tiominv / no_nodes;
             t = 0.0;
             if (tiominv != 0.) t = 1.e0 / tiominv;
             tpc = 0.0;
             if (tmax != 0.) tpc = t * 100.0 / tmax;
             printf("\n BTIO -- statistics:\n");
             printf("   I/O timing in seconds   : %14.2f\n", t);
             printf("   I/O timing percentage   : %14.2f\n", tpc);
             printf("   Total data written (MB) : %14.2f\n", mbytes);
             printf("   I/O data rate  (MB/sec) : %14.2f\n", mbytes*tiominv);
          }

         c_print_results("BT", class, grid_points(1), 
           grid_points(2), grid_points(3), niter, /*MAXCELLS*MAXCELLS,*/
           total_nodes, tmax, mflops, "          floating point", 
           verified, NPBVERSION,COMPILETIME, NPB_CS1, NPB_CS2, NPB_CS3, NPB_CS4, NPB_CS5, 
           NPB_CS6, "(none)");

       }

 done:
       upc_global_exit(0);

       return 0;

}

