
//---------------------------------------------------------------------
//---------------------------------------------------------------------
#include <stdio.h>
#include <unistd.h>
#include "header.h"
#include "upcnpb.h"
#include <upc_io.h>

static upc_file_t *btio_fp;
static void acc_sub_norms(int idump_cur);

upc_off_t iseek;

static int rwcall_sub = 0;


void setup_btio(void) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

      int m;
      int cio, count, max_count;

      if (node == root) {
         /* remove old file if exists */
         unlink(filenm);
      }
      upc_barrier;

      btio_fp = upc_all_fopen(filenm, 
                              UPC_INDIVIDUAL_FP|UPC_RDWR|UPC_CREATE, 
                              0, NULL);

      if (!btio_fp) {
          printf(" Error opening output file\n");
          upc_global_exit(1);
      }

      /* to track differences in the number of upc_all_fread/fwrite calls
         among all threads */
      count = 0;
      for (cio = 1; cio <= ncells; cio++) {
         count += cell_size(2,cio) * cell_size(3,cio);
      }
      mupc_reduce_i4(&count, &max_count, 1, SMP_OPR_MAX, -1);
      rwcall_sub = max_count - count;
//      printf("node %d, rwcall_sub = %d\n", node, rwcall_sub);

      for (m = 1; m <= 5; m++) {
         xce_sub(m) = 0.e0;
      }

      idump_sub = 0;

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

void output_timestep(void) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

      int jio, kio, cio, ix;
      upc_off_t offset;
      size_t count;

      for (cio = 1; cio <= ncells; cio++) {
          for (kio = 0; kio <= cell_size(3,cio)-1; kio++) {
              for (jio = 0; jio <= cell_size(2,cio)-1; jio++) {
                  iseek=5*(cell_low(1,cio) +
                         PROBLEM_SIZE*((cell_low(2,cio)+jio) +
                         PROBLEM_SIZE*((cell_low(3,cio)+kio) +
                         PROBLEM_SIZE*idump_sub)));

                  count=5*cell_size(1,cio);

		  offset = upc_all_fseek(btio_fp, iseek*sizeof(double), UPC_SEEK_SET);
                  if (offset == -1) {
		     printf("Error seeking to correct position in file\n");
		     upc_global_exit(-1);
		  }

		  offset = upc_all_fwrite_local(btio_fp, &u(1,0,jio,kio,cio),
						sizeof(double), count, 0);
                  if (offset == -1) {
                      printf(" Error writing to file\n");
                      upc_global_exit(-1);
                  }
              }
          }
      }

      /* perform a few null i/o to catch up with the rest */
      count = 0;
      for (cio = 0; cio < rwcall_sub; cio++) {
	 offset = upc_all_fseek(btio_fp, 0, UPC_SEEK_CUR);
	 offset = upc_all_fwrite_local(btio_fp, &u(1,0,0,0,1), 0, count, 0);
      }

      idump_sub = idump_sub + 1;
      if (rd_interval > 0) {
         if (idump_sub >= rd_interval) {

            acc_sub_norms(idump+1);

            idump_sub = 0;
         }
      }

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

static void acc_sub_norms(int idump_cur) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

      int jio, kio, cio, ix, ii, m, ichunk;
      upc_off_t offset;
      size_t count;

      double xce_single[5];

      ichunk = idump_cur - idump_sub + 1;
      for (ii = 0; ii <= idump_sub-1; ii++) {
        for (cio = 1; cio <= ncells; cio++) {
          for (kio = 0; kio <= cell_size(3,cio)-1; kio++) {
              for (jio = 0; jio <= cell_size(2,cio)-1; jio++) {
                  iseek=5*(cell_low(1,cio) +
                         PROBLEM_SIZE*((cell_low(2,cio)+jio) +
                         PROBLEM_SIZE*((cell_low(3,cio)+kio) +
                         PROBLEM_SIZE*ii)));

                  count=5*cell_size(1,cio);

		  offset = upc_all_fseek(btio_fp, iseek*sizeof(double), UPC_SEEK_SET);
		  if (offset == -1) {
		     printf("Error seeking to right position in file\n");
		     upc_global_exit(-1);
		  }

		  offset = upc_all_fread_local(btio_fp, &u(1,0,jio,kio,cio),
					       sizeof(double), count, 0);
                  if (offset == -1) {
                      printf(" Error reading back file\n");
                      upc_global_exit(-1);
                  }
              }
          }
        }

        /* perform a few null i/o to catch up with the rest */
        count = 0;
        for (cio = 0; cio < rwcall_sub; cio++) {
	   offset = upc_all_fseek(btio_fp, 0, UPC_SEEK_CUR);
	   offset = upc_all_fread_local(btio_fp, &u(1,0,0,0,1), 0, count, 0);
        }

        if (node == root) printf(" Reading data set %4d\n", ii+ichunk);

        error_norm(xce_single);
        for (m = 1; m <= 5; m++) {
           xce_sub(m) = xce_sub(m) + xce_single[m-1];
        }
      }

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

void btio_cleanup(void) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

      if (upc_all_fclose(btio_fp) == -1) {
         printf("Error closing file\n");
         upc_global_exit(-1);
      }

      return;
}

//---------------------------------------------------------------------
//---------------------------------------------------------------------

void accumulate_norms(double xce_acc[]) {

//---------------------------------------------------------------------
//---------------------------------------------------------------------

      int m;

      if (rd_interval <= 0) {

      btio_fp = upc_all_fopen(filenm, UPC_INDIVIDUAL_FP|UPC_RDONLY, 0, NULL);
      if (!btio_fp) {
         printf("Error opening file for read\n");
         upc_global_exit(1);
      }

//     clear the last time step

      clear_timestep();

//     read back the time steps and accumulate norms

      acc_sub_norms(idump);

      upc_all_fclose(btio_fp);
      }

      for (m = 1; m <= 5; m++) {
         xce_acc[m-1] = xce_sub(m) / (double)(idump);
      }

      return;
}

