

subroutine mpi_openmp_shifter(dim,mpi_zeon, mpi_zeon0,memax, safety_extra_bounds,&
                              mpi_rank, left_rank_MPI, right_rank_MPI,nr_repeats,repeat_cnt,&
                              kzeon,iright,ileft,toroidal_comm,nproc_toroidal,myrank_partd)                       
                              
  implicit none
  
  include 'mpif.h'
  
  integer :: left_rank_MPI, right_rank_MPI, buffer_size,zeon_it,myrank_partd
  integer :: safety_extra_bounds, mpi_rank, time_levels,toroidal_comm,nproc_toroidal
  integer :: memax, m, m0, me, i, j, dest, mtop, lasth, outer_loop_cnt,dim
  integer :: omp_get_num_threads, omp_get_thread_num, nthreads, thread_id
  integer :: nr_of_holes_can_be_filled, already_filled_holes, temp_send_zeon_cnt
  integer :: mrecv, send_zeon_cnt, mrecvleft, mrecvright
  integer :: istatus(MPI_STATUS_SIZE), ierr, idest, isource, isendtag, irecvtag
  integer :: isendcount, irecvcount, nr_repeats,repeat_cnt
  integer :: msendleft(2),msendright(2)
  integer, parameter :: mpi_Rsize=MPI_DOUBLE_PRECISION
  
  real(8),allocatable :: sendright(:,:)
  real(8),allocatable :: sendleft(:,:)
  real(8),allocatable :: recvleft(:,:)
  real(8),allocatable :: recvright(:,:)
  
  ! stores zeon position of each sent particle
  integer,dimension(1:memax) :: kzeon,iright,ileft
  
  real(8),dimension(dim,memax+safety_extra_bounds) :: mpi_zeon,mpi_zeon0
  
  
  if(repeat_cnt==1 .AND. mpi_rank==0 .AND. myrank_partd == 0) write(*,*) ' '
  if(repeat_cnt==1 .AND. mpi_rank==0 .AND. myrank_partd == 0) write(*,*) '    --------------------- '
  if(repeat_cnt==1 .AND. mpi_rank==0 .AND. myrank_partd == 0) write(*,*) ' '
  
  if(repeat_cnt==1 .AND. mpi_rank==0 .AND. myrank_partd == 0) then
    write(*,*) '  -> MPI Benchmark starts: myrank_toroidal=',mpi_rank,'myrank_partd=',myrank_partd,' left_rank_MPI=',&
      left_rank_MPI,' right_rank_MPI=',right_rank_MPI, 'memax=', memax
    write(*,*) ' '
   !write(*,*) 'mpi_zeon on MPI Process ', mpi_rank, '='
   !write(*,*) mpi_zeon
  endif

  m0=1
  me = memax
  outer_loop_cnt = 0
  send_zeon_cnt = 0
  
  100 outer_loop_cnt=outer_loop_cnt+1 ! we'll have two rounds: first round for "2", second round for "7"
  
    kzeon = 0
    send_zeon_cnt = 0
    msendleft=0
    msendright=0

    !  ******************* PARTICLE WORK (which ones will be sent) *******************
    do m = m0, me    

      dest = int(mpi_zeon(1,m))
        
      if( (dest >= 0) .AND. (dest .NE. mpi_rank) ) then !means that this particle will be shifted
      
        kzeon(send_zeon_cnt + 1) = m
        send_zeon_cnt = send_zeon_cnt + 1
        !if(mpi_rank==0) write(*,*) 'Z send_zeon_cnt=',send_zeon_cnt, 'dest=', dest
        
        ! particle move to the right:
        if( Mod(dest - mpi_rank + nproc_toroidal, nproc_toroidal) <= nproc_toroidal/2 ) then
          msendright(1)=msendright(1)+1
          iright(msendright(1))=m
        else
          msendleft(1)=msendleft(1)+1
          ileft(msendleft(1))=m  
        endif
      
      endif
    enddo
    
    
    !  ******************* ALLREDUCE *******************
    if(outer_loop_cnt > 1) then
 
      ! total # of particles to be shifted
      mrecv=0
   
      call MPI_ALLREDUCE(send_zeon_cnt,mrecv,1,MPI_INTEGER,MPI_SUM,toroidal_comm,ierr)

      ! no particle to be shifted, return
      if ( mrecv == 0 ) then
        repeat_cnt = repeat_cnt + 1
        memax=me
        return
      endif
    endif


    !  ******************* PACK PARTICLES *******************
    ! an extra space to prevent zero size when send_zeon_cnt=0
    allocate( sendright(1:(dim*2),1:max(1,msendright(1) ) ), sendleft(1:(dim*2),1:max(1,msendleft(1) ) ) )

    ! pack particle to move right
    !$omp parallel do private(i)
    do i=1,msendright(1)
      sendright(1:dim,i) = mpi_zeon(:,iright(i))
      sendright((dim+1):(dim*2),i) = mpi_zeon0(:,iright(i))
    enddo
    
    ! pack particle to move left
    !$omp parallel do private(i)
    do i=1,msendleft(1)
      sendleft(1:dim,i) = mpi_zeon(:,ileft(i))
      sendleft((dim+1):(dim*2),i) = mpi_zeon0(:,ileft(i))
    enddo
 

    ! ******************* FILL THE HOLES *******************
    mtop=me
! # of particles remain on local PE
    me=me-msendleft(1)-msendright(1)

    lasth=send_zeon_cnt
    do i=1,send_zeon_cnt
      m=kzeon(i)
      if(m>me) exit  !Break out of the DO loop
      do while(mtop == kzeon(lasth))
        mtop=mtop-1
        lasth=lasth-1
      enddo

      mpi_zeon(:,m)=mpi_zeon(:,mtop)
      mpi_zeon0(:,m)=mpi_zeon0(:,mtop)
      mtop=mtop-1
    enddo

    ! ******************* SEND/RECV PARTICLES *******************
    ! send # of particle to move right
    mrecvleft=0
    idest=right_rank_MPI
    isource=left_rank_MPI
    isendtag=0
    irecvtag=0
  
    call MPI_SENDRECV(msendright,1,MPI_INTEGER,idest,isendtag,&
         mrecvleft,1,MPI_INTEGER,isource,irecvtag,toroidal_comm,istatus,ierr)

    allocate(recvleft(dim*2,max(1,mrecvleft)))
 
    ! send particle to right and receive from left
    recvleft=0
    isendcount=max(1,msendright(1))*dim*2
    irecvcount=max(1,mrecvleft)*dim*2
    call MPI_SENDRECV(sendright,isendcount,mpi_Rsize,idest,isendtag,recvleft,&
         irecvcount,mpi_Rsize,isource,irecvtag,toroidal_comm,istatus,ierr)
         
! ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    ! send # of particle to move left
    mrecvright=0
    idest=left_rank_MPI
    isource=right_rank_MPI
    isendtag=0
    irecvtag=0
  
    call MPI_SENDRECV(msendleft,1,MPI_INTEGER,idest,isendtag,&
         mrecvright,1,MPI_INTEGER,isource,irecvtag,toroidal_comm,istatus,ierr)

    allocate(recvright(dim*2,max(1,mrecvright)))
 
    ! send particle to left and receive from right
    recvright=0
    isendcount=max(1,msendleft(1))*dim*2
    irecvcount=max(1,mrecvright)*dim*2
    call MPI_SENDRECV(sendleft,isendcount,mpi_Rsize,idest,isendtag,recvright,&
         irecvcount,mpi_Rsize,isource,irecvtag,toroidal_comm,istatus,ierr)     
         


    ! ******************* UNPACK PARTICLES (adding them to zeon) *******************
    !$omp parallel do private(m)
    do m=1,mrecvleft
      mpi_zeon(:,m+me)=recvleft(1:dim,m)
      mpi_zeon0(:,m+me)=recvleft((dim+1):(2*dim),m)
    enddo

    !$omp parallel do private(m)
    do m=1,mrecvright
      mpi_zeon(:,m+me+mrecvleft)=recvright(1:dim,m)
      mpi_zeon0(:,m+me+mrecvleft)=recvright((dim+1):(2*dim),m)
    enddo

    ! current partcle number= original - sent + received
    me=me+mrecvleft+mrecvright
    deallocate(sendright,sendleft,recvleft,recvright)
    m0=me-mrecvright-mrecvleft+1  ! starting index of received particles
    
 goto 100 ! outer_loop_cnt
  

end subroutine mpi_openmp_shifter  
