module mod_reduction
  implicit none
  private
  public :: caf_reduce, caf_prefix_reduce
  integer, parameter :: dk = kind(1.0d0)
  abstract interface
     real(dk) function rf(a, b)
        import :: dk
        real(dk), intent(in) :: a, b
     end function
  end interface
  real(kind=dk) :: g[*]
contains
  real(kind=dk) function caf_reduce(x, ufun)
    real(kind=dk), intent(in) :: x
    procedure(rf) :: ufun

    real(kind=dk) :: y 

    if (this_image() == 1) then
       g = x
       sync images(*)
    else
       sync images(1)
       critical
          y = g[1]
          g[1] = ufun(x,y) 
       end critical
    end if
    sync all
    caf_reduce = g[1]
    sync all ! protect against subsequent write of g
  end function caf_reduce
  real(kind=dk) function caf_prefix_reduce(x, ufun)
    real(kind=dk), intent(in) :: x
    procedure(rf) :: ufun

    real(kind=dk) :: y 
    integer :: me

    me = this_image()
    if (me == 1) then 
       g = x
       caf_prefix_reduce = x
    else
       sync images ( (/ me, me-1 /) )
       y = g[me-1]
       g = ufun(x,y)
       caf_prefix_reduce = g
    end if
    if (me < num_images()) sync images ( (/ me, me+1 /) )
    sync all ! protect against subsequent write of g on 1
  end function caf_prefix_reduce
end module
