program triangular
  implicit none
! a is the data structure used in this program; a_serial was used
! in the serial program
! a_serial(i)  =  a(i/nprocs)[mod(i,nprocs)]  
!                                             i = 1..n
! a_serial(me+(i_local-1)*nprocs) = a(i_local)[me]  
!                                             i_local = 1..rows_per_proc
!                                             me = 1..nprocs 
!
  type :: tri_matrix
    real, allocatable :: row(:)
  end type
  type(tri_matrix), allocatable :: a(:)[:]

  integer :: n[*], i_row[*] 
  integer ::  nproc, me, rows_per_proc
  integer ::  i, j, i_local, n_elem

  me = this_image()
  if (me == 1) then
     read(*, *) n, i_row
  end if   
  sync all           ! Now, n[1] and i_row[1] are accessible from each image.
  n = n[1]           ! Copying of information into the local portions of
  i_row = i_row[1]   ! variables n and i_row on each image.
  
  nproc = num_images()
  
  rows_per_proc = n / nproc
  if (mod(n, nproc) > 0) rows_per_proc = rows_per_proc + 1
  allocate(a(rows_per_proc)[*])
!
! create triangular matrix :
  i_local = 1
  n_elem = 0
  do i = me, n, nproc
    allocate(a(i_local)%row(n - i + 1))
    n_elem = n_elem + n - i + 1
    i_local = i_local + 1
  end do
!
! initialize matrix A(i, j) = i + j  
  i_local = 1
  do i = me, n, nproc
    do j = 1, n - i + 1
       a(i_local)%row(j) = real(i) + real(j)
    end do
    i_local = i_local + 1
  end do
!
! write row number i_row to standard output
  sync all
  i_local = (i_row - 1) / nproc + 1
  write(*, fmt='(('' Row '',i0,'' on image '',i0,'':''),1000(1X,F4.1))') &
        i_row, me, (a(i_local)[mod(i_row-1,nproc)+1]%row(j), &
        j=1, n - i_row + 1)

  write(*, fmt='(''Number of elements on image '',i0,'': '',i0)') &
           me, n_elem
!  
! clean up
  sync all
! the following three lines should not be needed
! (workaround for a bug in some compilers)
  do i=1, size(a)
    if (allocated(a(i)%row)) deallocate(a(i)%row)
  end do
  deallocate(a)  ! implies deallocation of components
end program
